/*
 * title: search.c
 * summary: search torsions file for a pattern in backbone angles
 * date:    May 4, 2004
 * author:  Nick Fitzkee (nickfitzkee@jhu.edu)
 *
 */

#include "search.h"

int main(int argc, char** argv)
{
  int nfiles=0, i, j;
  int vec[FRG_LEN] = {0};
  FILE* inf;
  FILE* outf;
  FILE* vf = NULL;
  char line[MAX_BUFSZ];
  hit_t* hits;
  unsigned long searched=0;
  
  int ntypes = ipow(SS_TYPES, FRG_LEN);

  hits = malloc(sizeof(unsigned long)*ntypes);
  for (i=0; i<ntypes; i++) hits[i] = 0;

  if (argc < 3 || argc > 4) usage();
  
  inf = fopen(argv[1], "r");

  if (argc == 4) {
    vf = fopen(argv[3], "w");
  }

  if (inf == NULL) {
    printf("Unable to open input file: %s\n", argv[1]);
    return 1;
  }

  printf("Searching for fragments of size %i (%i ss types)...\n",
	 FRG_LEN, SS_TYPES);

  while(fgets(line, MAX_BUFSZ, inf)) {
    if (line[0] == '\0' || line[0] == '#') continue;
    str_strip(line);

    /*printf("Searching file: %s\n", line);*/
    if (search_one_file(line, vf, hits, &searched))
      return 1;
    nfiles++;
  }

  printf("%i files examined.\n\n", nfiles);
  if (argc == 3) fclose(vf);



  printf("Writing output file... ");
  outf = fopen(argv[2], "w");

  i = 0;
  do {
    for (j=0; j<FRG_LEN; j++)
      fprintf(outf, "%c", ss_elements[vec[j]].letter);

    fprintf(outf, " %10lu %10lu %8.3e\n", hits[i], searched, 
	    ((double)hits[i])/((double)searched));
    i++;
  } while(inc_vector(vec, FRG_LEN, SS_TYPES));
  printf("done!\n");

  free(hits);
  fclose(outf);
  fclose(inf);
  return 0;
}



/* Search-Related Functions **************************************************/

/* search_one_file
 *
 * Takes a file name from a torsion file list, uncompresses it, loads it into
 * memory, and performs the search of all secondary structure enumerations.
 * If all is well, it returns zero, if there is an error, it returns a
 * negative number corresponding to the (file read) error it received.  It
 * increments hits and searched according to the number of fragments
 * that matched.
 *
 */

int search_one_file(char* tf, FILE* output, hit_t* hits, 
		    long unsigned* searched)
{
  long residx[MAX_RES];
  char rescod[MAX_RES];
  struct tors_t tors [MAX_RES];
  char pdb[6], sticd;
  int start, len, nres, i, idxst, idxen;

  int errnum = load_file(tf, residx, rescod, tors, &start, &sticd,
			 &len, &nres, pdb);
  if (errnum) return errnum;

  for (idxst=0; 
       !(residx[idxst]==start && rescod[idxst]==sticd) && idxst<nres; 
       idxst++) {}
  idxen = idxst + len;

  /*
  for (i=0; i<nres; i++)
    printf("%5li %10.4f %10.4f %10.4f\n", residx[i], 
	   tors[i].phi, tors[i].psi, tors[i].ome);
  */

  for (i = idxst; i <= idxst+len-FRG_LEN; i++) {
    /* printf("start from: %5li\n", residx[i]); */
    search_one_seq(i, residx, rescod, tors, hits, pdb, output);
    (*searched)++;
  }

  return 0;
}

void search_one_seq(int i, long residx [], char rescod [], 
		    struct tors_t tors[], hit_t* hits, char* pdb, 
		    FILE* output)
{
  int result = 0;
  int end = i + FRG_LEN;

  if (output == NULL) {

    for (; i<end; i++) {
      if (is_helix(tors[i].phi, tors[i].psi, tors[i].ome))
	result = result * SS_TYPES + ss_helix;
      else if (is_strand(tors[i].phi, tors[i].psi, tors[i].ome)) 
	result = result * SS_TYPES + ss_strand;
      else if (is_pp2(tors[i].phi, tors[i].psi, tors[i].ome))
	result = result * SS_TYPES + ss_pp2;
      else if (is_lhh(tors[i].phi, tors[i].psi, tors[i].ome))
	result = result * SS_TYPES + ss_lhh;
      else
	result = result * SS_TYPES + ss_other;
    }
  }
   
  else {
    fprintf(output, "%5s %5li %c ", pdb, residx[i], rescod[i]);

    for (; i<end; i++) {
      if (is_helix(tors[i].phi, tors[i].psi, tors[i].ome)) {
	fprintf(output, "%c", ss_elements[ss_helix].letter);
	result = result * SS_TYPES + ss_helix;
      }
      else if (is_strand(tors[i].phi, tors[i].psi, tors[i].ome)) {
	fprintf(output, "%c", ss_elements[ss_strand].letter);
	result = result * SS_TYPES + ss_strand;
      }
      else if (is_pp2(tors[i].phi, tors[i].psi, tors[i].ome)) {
	fprintf(output, "%c", ss_elements[ss_pp2].letter);
	result = result * SS_TYPES + ss_pp2;
      }
      else if (is_lhh(tors[i].phi, tors[i].psi, tors[i].ome)) {
	fprintf(output, "%c", ss_elements[ss_lhh].letter);
	result = result * SS_TYPES + ss_lhh;
      }
      else {
	fprintf(output, "%c", ss_elements[ss_other].letter);
	result = result * SS_TYPES + ss_other;
      }
    }
    fprintf(output, "\n");
  }

  hits[result]++;
  return;
}


/* load_file
 *
 * Given a file name (tf) this function loads relevant torsion information
 * from that file into the given index and torsion angle structures, 
 * idx and torsions.  Additionally, the size and starting position of the
 * coil unit is returned, along with the number of residues in the fragment.
 *
 */

int load_file(char *tf, long idx[], char cod[], struct tors_t torsions[], 
	      int *start, char* stcod, int* len, int* nres, char* pdb)
{
  char line[MAX_BUFSZ];
  char* result;
  gzFile gzf = gzopen(tf, "r");
  int i;

  if (gzf == NULL) {
    printf("Unable to load torsion file: %s\n", tf);
    return -1;
  }
  gzgets(gzf, line, MAX_BUFSZ);

  if (!strlen(line)) {
    printf("Empty torsion file: %s\n", tf);
    return -1;
  }

  if (atoi(line) != 1) {
    printf("Invalid torsion file: %s\n", tf);
    return -1;
  }

  gzgets(gzf, line, MAX_BUFSZ);

  if (!strlen(line)) {
    printf("Invalid torsion file: %s\n", tf);
    return -1;
  }

  *nres = atoi(line+2);

  if (*nres > MAX_RES) {
    printf("Too many residues in torsion file: %s\n", tf);
    return -1;
  }

  for (i=0; i< *nres; i++) {
    result = gzgets(gzf, line, MAX_BUFSZ);
    idx[i] = strtol(line, NULL, 10);
    cod[i] = line[9];
    torsions[i].phi = strtod(line+16, NULL);
    torsions[i].psi = strtod(line+27, NULL);
    torsions[i].ome = strtod(line+38, NULL);
  }

  /* Parse the file name--assume extension is .xxx.gz */

  *stcod = tf[strlen(tf)-8];
  if (*stcod == '_') *stcod = ' ';

  for (i=strlen(tf)-8; i >= 0; i--) {
    if (tf[i] == '.') {
      *start = strtod(tf+i+1, NULL);
      tf[i] = '\0';
      break;
    }
  }

  i--;
  for (; i >= 0; i--) {
    if (tf[i] == '.') {
      *len = strtod(tf+i+1, NULL);
      break;
    }
  }

  if (i==0) {
    printf("Unable to parse file name: %s\n", tf);
    return -1;
  }

  strncpy(pdb, tf+i-5, 5);
  pdb[5] = '\0';

  gzclose(gzf);
  return 0;
}

/* Utility Functions *********************************************************/


void usage()
{
  printf("usage: search <torfile list> <output> [<verbose file>]\n");
  exit(0);
}


/*
 * inc_vector
 *
 * Given a vector vec, of dimensionality dim, this function increments
 * the least significant digit with carryover modulo mod.  The vector
 * is modified in place.  If there is an overflow after calculation,
 * the fucntion evaluates to "false," otherwise it's true.
 *
 */

int inc_vector(unsigned int vec[], int dim, unsigned int mod)
{
  for (dim--; dim >= 0; dim--) {
    vec[dim] = (vec[dim] + 1) % mod;
    if (vec[dim]) break;
  }

  return !(dim < 0);
}


/* 
 * str_strip
 *
 * Given a string, this function searches for whitespace at the beginning
 * and end of the line and removes it, if present.  Algorithm taken from
 * from Eric Youngdale.
 *
 */

#define white(c) (c == ' ' || c == '\t' || c == '\n')

void str_strip (char *string)
{
  int i = 0;

  while (white(string[i]))
    i++;

  if (i) strcpy (string, string + i);
  i = strlen (string) - 1;

  while (i > 0 && white(string[i]))
    i--;

  string[++i] = '\0';
}

/* ipow
 *
 * calculate the integer power of i**n.  doesn't use any tricky math 
 * speedups.
 *
 */

long int ipow(int i, unsigned int n) 
{
  int j = 0, result = 1;

  for (; j<n; j++)
    result *= i;

  return result;
}

/* Backbone Analysis Functions ***********************************************/

static inline int is_helix(double phi, double psi, double ome)
{
  return (phi >=  -75.0 &&
	  phi <=  -45.0 &&
	  psi >=  -60.0 &&
	  psi <=  -30.0);
}

static inline int is_strand(double phi, double psi, double ome)
{
  return (phi >= -135.0 &&
	  phi <= -105.0 &&
	  psi >=  120.0 &&
	  psi <=  150.0);
}

static inline int is_pp2(double phi, double psi, double ome)
{
  return (phi >=  -80.0 &&
	  phi <=  -55.0 &&
	  psi >=  130.0 &&
	  psi <=  155.0);
}

static inline int is_lhh(double phi, double psi, double ome)
{
  return (phi >= 45.0 &&
	  phi <= 75.0 &&
	  psi >=  0.0 &&
	  psi <= 60.0);
}

static inline int is_other(double phi, double psi, double ome)
{
  return 1;
}
