/*
 * title: search.h
 * summary: definitions, prototypes, and enumerations for search.c
 * date:    May 5, 2004
 * author:  Nicholas Fitzkee (nickfitzkee@jhu.edu)
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <zlib.h>
#include <math.h>


/*
 * FRG_LEN defines the fragment length that will be examined in the search.
 * It should be a reasonable number--short enough to capture information
 * we wish to see, but long enough to describe real units in proteins.
 *
 */

#define FRG_LEN 5

/*
 * Information about the number and type of backbone torsion classifications.
 * The number of SS_TYPES should match the actual number, and the enumeration
 * list and corresponding struct should pair the types with their actual
 * functions.
 *
 */

#define SS_TYPES 5

/*
 * The maximum number of residues expected in a given torsions file.
 *
 */

#define MAX_RES 150

/*
 * The maximum line length for reading file buffers.
 *
 */

#define MAX_BUFSZ 150

/* Backbone Analysis Functions */

static inline int is_helix(double, double, double);
static inline int is_strand(double, double, double);
static inline int is_pp2(double, double, double);
static inline int is_lhh(double, double, double);
static inline int is_other(double, double, double);


/* Enumeration of SS types */

enum ss_enum {
  ss_helix = 0,
  ss_strand,
  ss_pp2,
  ss_lhh,
  ss_other
};

/* Struct that links enumeration, text, and function */

struct ss_entry {
  enum ss_enum type;
  char letter;
  void* tfunc;
};

static struct ss_entry ss_elements[] = {
  {ss_helix,  'H', &is_helix},
  {ss_strand, 'E', &is_strand},
  {ss_pp2,    'P', &is_pp2},
  {ss_lhh,    'L', &is_lhh},
  {ss_other,  'O', &is_other},
  {0xFF,      0,   NULL}
};

/* 
 * Other structs, etc. that are used in the code.
 */

typedef unsigned long hit_t;

struct tors_t {
  double phi, psi, ome;
};

/* Function Prototypes *******************************************************/

/* Utility Functions */
void usage();
int inc_vector(unsigned int [], int, unsigned int);
void str_strip (char *);
long int ipow(int, unsigned int);

/* Search-Related Functions */
int search_one_file (char *, FILE*, hit_t*, unsigned long*);
int load_file(char*, long [], char [], struct tors_t [], 
	      int*, char*, int*, int*, char*);
void search_one_seq(int, long [], char [], struct tors_t [], 
		    hit_t*, char*, FILE*);
